# MIT License
#
# Copyright (c) 2018 Tom Runia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Aaron O'Leary (dev@aaren.me)
# Date Created: 2016-02-28

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import scipy
import scipy.signal
import scipy.optimize
import scipy.special
from scipy.special import factorial

__all__ = ['Morlet', 'Paul', 'DOG', 'Ricker', 'Marr', 'Mexican_hat']


class Morlet(object):
    def __init__(self, w0=6):
        """w0 is the nondimensional frequency constant. If this is
        set too low then the wavelet does not sample very well: a
        value over 5 should be ok; Terrence and Compo set it to 6.
        """
        self.w0 = w0
        if w0 == 6:
            # value of C_d from TC98
            self.C_d = 0.776

    def __call__(self, *args, **kwargs):
        return self.time(*args, **kwargs)

    def time(self, t, s=1.0, complete=True):
        """
        Complex Morlet wavelet, centred at zero.

        Parameters
        ----------
        t : float
            Time. If s is not specified, this can be used as the
            non-dimensional time t/s.
        s : float
            Scaling factor. Default is 1.
        complete : bool
            Whether to use the complete or the standard version.

        Returns
        -------
        out : complex
            Value of the Morlet wavelet at the given time

        See Also
        --------
        scipy.signal.gausspulse

        Notes
        -----
        The standard version::

            pi**-0.25 * exp(1j*w*x) * exp(-0.5*(x**2))

        This commonly used wavelet is often referred to simply as the
        Morlet wavelet.  Note that this simplified version can cause
        admissibility problems at low values of `w`.

        The complete version::

            pi**-0.25 * (exp(1j*w*x) - exp(-0.5*(w**2))) * exp(-0.5*(x**2))

        The complete version of the Morlet wavelet, with a correction
        term to improve admissibility. For `w` greater than 5, the
        correction term is negligible.

        Note that the energy of the return wavelet is not normalised
        according to `s`.

        The fundamental frequency of this wavelet in Hz is given
        by ``f = 2*s*w*r / M`` where r is the sampling rate.

        """
        w = self.w0

        x = t / s

        output = np.exp(1j * w * x)

        if complete:
            output -= np.exp(-0.5 * (w ** 2))

        output *= np.exp(-0.5 * (x ** 2)) * np.pi ** (-0.25)

        return output

    # Fourier wavelengths
    def fourier_period(self, s):
        """Equivalent Fourier period of Morlet"""
        return 4 * np.pi * s / (self.w0 + (2 + self.w0 ** 2) ** .5)

    def scale_from_period(self, period):
        """
        Compute the scale from the fourier period.
        Returns the scale
        """
        # Solve 4 * np.pi * scale / (w0 + (2 + w0 ** 2) ** .5)
        #  for s to obtain this formula
        coeff = np.sqrt(self.w0 * self.w0 + 2)
        return (period * (coeff + self.w0)) / (4. * np.pi)

    # Frequency representation
    def frequency(self, w, s=1.0):
        """Frequency representation of Morlet.

        Parameters
        ----------
        w : float
            Angular frequency. If `s` is not specified, i.e. set to 1,
            this can be used as the non-dimensional angular
            frequency w * s.
        s : float
            Scaling factor. Default is 1.

        Returns
        -------
        out : complex
            Value of the Morlet wavelet at the given frequency
        """
        x = w * s
        # Heaviside mock
        Hw = np.array(w)
        Hw[w <= 0] = 0
        Hw[w > 0] = 1
        return np.pi ** -.25 * Hw * np.exp((-(x - self.w0) ** 2) / 2)

    def coi(self, s):
        """The e folding time for the autocorrelation of wavelet
        power at each scale, i.e. the timescale over which an edge
        effect decays by a factor of 1/e^2.

        This can be worked out analytically by solving

            |Y_0(T)|^2 / |Y_0(0)|^2 = 1 / e^2
        """
        return 2 ** .5 * s


class Paul(object):
    def __init__(self, m=4):
        """Initialise a Paul wavelet function of order `m`.
        """
        self.m = m

    def __call__(self, *args, **kwargs):
        return self.time(*args, **kwargs)

    def time(self, t, s=1.0):
        """
        Complex Paul wavelet, centred at zero.

        Parameters
        ----------
        t : float
            Time. If `s` is not specified, i.e. set to 1, this can be
            used as the non-dimensional time t/s.
        s : float
            Scaling factor. Default is 1.

        Returns
        -------
        out : complex
            Value of the Paul wavelet at the given time

        The Paul wavelet is defined (in time) as::

            (2 ** m * i ** m * m!) / (pi * (2 * m)!) \
                    * (1 - i * t / s) ** -(m + 1)

        """
        m = self.m
        x = t / s

        const = (2 ** m * 1j ** m * factorial(m)) \
            / (np.pi * factorial(2 * m)) ** .5
        functional_form = (1 - 1j * x) ** -(m + 1)

        output = const * functional_form

        return output

    # Fourier wavelengths
    def fourier_period(self, s):
        """Equivalent Fourier period of Paul"""
        return 4 * np.pi * s / (2 * self.m + 1)

    def scale_from_period(self, period):
        raise NotImplementedError()

    # Frequency representation
    def frequency(self, w, s=1.0):
        """Frequency representation of Paul.

        Parameters
        ----------
        w : float
            Angular frequency. If `s` is not specified, i.e. set to 1,
            this can be used as the non-dimensional angular
            frequency w * s.
        s : float
            Scaling factor. Default is 1.

        Returns
        -------
        out : complex
            Value of the Paul wavelet at the given frequency

        """
        m = self.m
        x = w * s
        # Heaviside mock
        Hw = 0.5 * (np.sign(x) + 1)

        # prefactor
        const = 2 ** m / (m * factorial(2 * m - 1)) ** .5

        functional_form = Hw * (x) ** m * np.exp(-x)

        output = const * functional_form

        return output

    def coi(self, s):
        """The e folding time for the autocorrelation of wavelet
        power at each scale, i.e. the timescale over which an edge
        effect decays by a factor of 1/e^2.

        This can be worked out analytically by solving

            |Y_0(T)|^2 / |Y_0(0)|^2 = 1 / e^2
        """
        return s / 2 ** .5


class DOG(object):
    def __init__(self, m=2):
        """Initialise a Derivative of Gaussian wavelet of order `m`."""
        if m == 2:
            # value of C_d from TC98
            self.C_d = 3.541
        elif m == 6:
            self.C_d = 1.966
        else:
            pass
        self.m = m

    def __call__(self, *args, **kwargs):
        return self.time(*args, **kwargs)

    def time(self, t, s=1.0):
        """
        Return a Derivative of Gaussian wavelet,

        When m = 2, this is also known as the "Mexican hat", "Marr"
        or "Ricker" wavelet.

        It models the function::

            ``A d^m/dx^m exp(-x^2 / 2)``,

        where ``A = (-1)^(m+1) / (gamma(m + 1/2))^.5``
        and   ``x = t / s``.

        Note that the energy of the return wavelet is not normalised
        according to `s`.

        Parameters
        ----------
        t : float
            Time. If `s` is not specified, this can be used as the
            non-dimensional time t/s.
        s : scalar
            Width parameter of the wavelet.

        Returns
        -------
        out : float
            Value of the DOG wavelet at the given time

        Notes
        -----
        The derivative of the Gaussian has a polynomial representation:

        from http://en.wikipedia.org/wiki/Gaussian_function:

        "Mathematically, the derivatives of the Gaussian function can be
        represented using Hermite functions. The n-th derivative of the
        Gaussian is the Gaussian function itself multiplied by the n-th
        Hermite polynomial, up to scale."

        http://en.wikipedia.org/wiki/Hermite_polynomial

        Here, we want the 'probabilists' Hermite polynomial (He_n),
        which is computed by scipy.special.hermitenorm

        """
        x = t / s
        m = self.m

        # compute the Hermite polynomial (used to evaluate the
        # derivative of a Gaussian)
        He_n = scipy.special.hermitenorm(m)
        gamma = scipy.special.gamma

        const = (-1) ** (m + 1) / gamma(m + 0.5) ** .5
        function = He_n(x) * np.exp(-x ** 2 / 2)

        return const * function

    def fourier_period(self, s):
        """Equivalent Fourier period of derivative of Gaussian"""
        return 2 * np.pi * s / (self.m + 0.5) ** .5

    def scale_from_period(self, period):
        raise NotImplementedError()

    def frequency(self, w, s=1.0):
        """Frequency representation of derivative of Gaussian.

        Parameters
        ----------
        w : float
            Angular frequency. If `s` is not specified, i.e. set to 1,
            this can be used as the non-dimensional angular
            frequency w * s.
        s : float
            Scaling factor. Default is 1.

        Returns
        -------
        out : complex
            Value of the derivative of Gaussian wavelet at the
            given time
        """
        m = self.m
        x = s * w
        gamma = scipy.special.gamma
        const = -1j ** m / gamma(m + 0.5) ** .5
        function = x ** m * np.exp(-x ** 2 / 2)
        return const * function

    def coi(self, s):
        """The e folding time for the autocorrelation of wavelet
        power at each scale, i.e. the timescale over which an edge
        effect decays by a factor of 1/e^2.

        This can be worked out analytically by solving

            |Y_0(T)|^2 / |Y_0(0)|^2 = 1 / e^2
        """
        return 2 ** .5 * s


class Ricker(DOG):
    def __init__(self):
        """The Ricker, aka Marr / Mexican Hat, wavelet is a
        derivative of Gaussian order 2.
        """
        DOG.__init__(self, m=2)
        # value of C_d from TC98
        self.C_d = 3.541


# aliases for DOG2
Marr = Ricker
Mexican_hat = Ricker


import torch
import torch.nn as nn

class MLP(nn.Module):
    def __init__(self, input_size = 513, n_class = 7, hid_1 = 256, hid_2 = 128):
        torch.manual_seed(28)
        super().__init__()
        self.input_size = input_size
        self.__type__ = 'mlp'
        self.__arch__ = 'mlp'
        self.__version__ = 'baseline'
        self.fc1 = nn.Linear(input_size,hid_1)
        self.fc2 = nn.Linear(hid_1,hid_2)
        self.fc3 = nn.Linear(hid_2, n_class)
        self.relu = nn.ReLU()
        self.dropout1 = nn.Dropout(0.5)
        self.dropout2 = nn.Dropout(0.3)
        self.init_weights = False

    def forward(self,x):
        x = self.relu(self.fc1(x))
        x = self.dropout1(x)
        x = self.relu(self.fc2(x))
        x = self.dropout2(x)
        x = self.fc3(x)
        x = nn.functional.log_softmax(x,-1)
        return x


# MIT License
# 
# Copyright (c) 2018 Tom Runia
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-04-16

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import numpy as np

from wavelets_pytorch.transform import WaveletTransform
from wavelets_pytorch.transform import WaveletTransformTorch

######################################

fps = 20
dt  = 1.0/fps
dj  = 0.125
unbias = False
t_min = 0

batch_sizes = np.asarray([1,8,16,32,64,128,256,512], np.int32)
durations = np.asarray([5,10,25,50,100], np.int32)
signal_lengths = durations*fps

num_runs = 5

runtimes_scipy = np.zeros((len(batch_sizes), len(signal_lengths), num_runs), np.float32)
runtimes_torch = np.zeros((len(batch_sizes), len(signal_lengths), num_runs), np.float32)

for batch_ind, batch_size in enumerate(batch_sizes):

    for length_ind, t_max in enumerate(durations):

        t = np.linspace(t_min, t_max, (t_max-t_min)*fps)
        print('#'*60)
        print('Benchmarking | BatchSize = {}, SignalLength = {}'.format(batch_size, signal_lengths[length_ind]))

        for run_ind in range(num_runs):

            random_frequencies = np.random.uniform(-0.5, 4.0, size=batch_size)
            batch = np.asarray([np.sin(2*np.pi*f*t) for f in random_frequencies])

            # Perform batch computation of SciPy implementation
            t_start = time.time()
            wa = WaveletTransform(dt, dj, unbias=unbias)
            power = wa.power(batch)
            runtimes_scipy[batch_ind,length_ind,run_ind] = time.time() - t_start
            #print("  Run {}/{} | SciPy: {:.2f}s".format(run_ind+1, num_runs, runtimes[batch_ind,length_ind,run_ind,0]))

            # Perform batch computation of Torch implementation
            t_start = time.time()
            wa = WaveletTransformTorch(dt, dj, unbias=unbias)
            power = wa.power(batch)
            runtimes_torch[batch_ind,length_ind,run_ind] = time.time() - t_start
            #print("  Run {}/{} | Torch: {:.2f}s".format(run_ind+1, num_runs, runtimes[batch_ind,length_ind,run_ind,1]))

        avg_scipy = np.mean(runtimes_scipy[batch_ind,length_ind,:])
        avg_torch = np.mean(runtimes_torch[batch_ind,length_ind,:])
        print('  Average SciPy: {:.2f}s'.format(avg_scipy))
        print('  Average Torch: {:.2f}s'.format(avg_torch))

np.save('./runtimes_scipy.npy', runtimes_scipy)
np.save('./runtimes_torch.npy', runtimes_torch)



# MIT License
#
# Copyright (c) 2018 Tom Runia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-04-16

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from wavelets_pytorch.transform import WaveletTransform        # SciPy version
from wavelets_pytorch.transform import WaveletTransformTorch   # PyTorch version

from wavelets_pytorch.wavelets import Morlet, Ricker

"""
Example script to demonstrate the CWT on a batch of random sinusoidal signals. 
We compare both the SciPy implementation and the PyTorch implementation. 
"""

dt = 0.1               # sampling frequency
dj = 0.125             # scale distribution parameter
batch_size = 32        # how many signals to process in parallel
cuda = True            # enable GPU

t = np.linspace(0., 10., int(10./dt))

# Both use a complex and real wavelet
for wavelet in [Morlet(), Ricker()]:

    # Sinusoidals with random frequency
    frequencies = np.random.uniform(-0.5, 2.0, size=batch_size)
    batch = np.asarray([np.sin(2*np.pi*f*t) for f in frequencies])

    # Initialize wavelet filter banks (scipy and torch implementation)
    wa_scipy = WaveletTransform(dt, dj, wavelet)
    wa_torch = WaveletTransformTorch(dt, dj, wavelet, cuda=cuda)

    # Performing wavelet transform (and compute scalogram)
    cwt_scipy = wa_scipy.cwt(batch)
    cwt_torch = wa_torch.cwt(batch)

    print(cwt_scipy.shape)
    print(cwt_torch.shape)

    # For plotting, see the examples/plot.py function.
    # ...


import json
import os 
import datetime
from easydict import EasyDict
import torch

def get_config_from_json(json_file):
    """
    Get the config from a json file
    :param json_file:
    :return: config(namespace) or config(dictionary)
    """
    # parse the configurations from the config json file provided
    with open(json_file, 'r') as config_file:
        config_dict = json.load(config_file)

    # convert the dictionary to a namespace using bunch lib
    config = EasyDict(config_dict)

    return config

def process_config(json_file):
    if not os.path.exists(json_file):
        try:
            config = get_config_from_json(f'./configs/{json_file}')
        except:
            raise RuntimeError(f'{json_file} not found')
    else: 
        config = get_config_from_json(json_file)
        
    time = str(datetime.datetime.now())[:-7]
    config.device = 'cuda' if torch.cuda.is_available() else 'cpu'
    config.timestamp = time
    return config 



import numpy as np
import librosa
import pywt 

from copy import deepcopy as dc
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly_resampler import FigureResampler, FigureWidgetResampler

from ..dataset_utils.datahelper import get_index


c = {'NP':'darkgoldenrod',    'C':'lightgreen',    'E1':'skyblue',    'E2':'deeppink',  'F':'crimson',   'G':'darkblue',   'pd':'olive'}

#######################
## Display waveforms ##
#######################
def get_position_label(waveform_indices, position):
    for wave_type in waveform_indices:
        for start, end in waveform_indices[wave_type]:
            if (start <= position) & (position <= end):
                return wave_type
            
def visualize_signal(recording, ana_file, sr = 100, ax = None, title = ''):

    plt.rcParams.update({'font.size': 14})
    '''
    Input:
        :param recording: wave signal of class np.ndarray
        :param ana: analysis file corresponding to the recording
        :param wave_type: 'whole' or str of {'NP', 'C', 'E1', 'E2', 'F', 'G', 'pd'}
        :param sr: sampling rate. Default = 100Hz.
    Output:
        plot of the full wave (wave types) with colors
    '''    
    
    time_axis = np.linspace(0, len(recording)/sr, len(recording))

    ana = dc(ana_file)
    ana['time'] = ana['time'].apply(lambda x: x*sr)
    ana['time'] = ana['time'].astype(int)
    waveform_indices = get_index(ana)
    
    custom_legends = []
    i = 0
    if ax == None:
        ax = plt.gca()
    for wave_type in waveform_indices.keys():
        for start, end in waveform_indices[wave_type]:
            
            ax.plot(time_axis[start:end+1],recording[start:end+1],color = c[wave_type])
        custom_legends.append(Line2D([0], [0], color=c[wave_type], lw=4))
        i+=1

    ax.legend(custom_legends, ['NP', 'C', 'pd', 'F', 'G','E1', 'E2'],loc = 'upper right', ncols=len(custom_legends), fontsize = 15)
    ax.set_xticks(np.arange(0,time_axis.max(),2000), labels=np.arange(0,time_axis.max(),2000), rotation = 45)
    ax.set_xlabel(f'Time (s). Sampling rate: {sr} (Hz)')
    ax.set_ylabel('Amplitude (V)')
    ax.set_title(title) if isinstance(title, str) else None

def visualize_waveform(recording, ana_file, waveform = None, in_between = False, padding = 512, seed: int = 100, idx = None, sr = 100, ax = None):

    plt.rcParams.update({'font.size': 14})
    '''
    Input:
        :param recording: wave signal of class np.ndarray
        :param ana: analysis file corresponding to the recording
        :param wave_type: 'whole' or str of {'NP', 'C', 'E1', 'E2', 'F', 'G', 'pd'}
        :param sr: sampling rate. Default = 100Hz.
    Output:
        plot of the full wave (wave types) with colors
    '''    
    if seed is not None: 
        np.random.seed(seed)


    time_axis = np.linspace(0,len(recording)/sr,len(recording))
    ana = dc(ana_file)
    ana['time'] = ana['time'].apply(lambda x: x*sr)
    ana['time'] = ana['time'].astype(int)
    waveform_indices = get_index(ana)

    if waveform is not None:
        if ax == None:
            ax = plt.gca()
        wave_type = waveform
        if idx is None:
            idx = np.random.choice(np.arange(len(waveform_indices[wave_type])))
        start, end = waveform_indices[wave_type][idx]
        # pad the wave with adjacent waveform to see the relationship
        ax.plot(time_axis[start:end], recording[start:end], color = c[wave_type], label = wave_type)
        ax.legend(loc = 'upper right')      

        if in_between == True:
            start_pad = start - padding
            end_pad = end + padding
            former_label = get_position_label(waveform_indices,start_pad)
            latter_label = get_position_label(waveform_indices,end_pad)
            if former_label != None:
                ax.plot(time_axis[start_pad:start+1],recording[start_pad:start+1],
                        color = c[former_label],label = former_label)

            if latter_label != None:
                ax.plot(time_axis[end-1:end_pad],recording[end-1:end_pad],
                        color = c[latter_label], label = latter_label)
    else:
        raise ValueError("Must be one of 'NP', 'C', 'E1', 'E2', 'F', 'G' or 'pd' ")
    return plt.gcf()

### Interactive
def interactive_visualization(recording, ana_file, smoothen = False, sr = 100, title = ''):
    time = np.linspace(0,len(recording)/sr,len(recording))
    ana = dc(ana_file)
    ana['time'] = ana['time'].apply(lambda x: x*sr)
    ana['time'] = ana['time'].astype(int)
    waveform_indices = get_index(ana)

    if smoothen == True:
        fig = FigureWidgetResampler(go.Figure())
    elif smoothen == False:
        fig = go.Figure()

    for wave in waveform_indices.keys():
        i = 0
        for start, end in waveform_indices[wave]:
            if i == 0:
                fig.add_trace(go.Scatter(x = time[start:end+1], y = recording[start:end+1],line=dict(color=c[wave]), mode = 'lines',
                                        legendgroup = wave, name = wave))   
            else:
                fig.add_trace(go.Scatter(x = time[start:end+1], y = recording[start:end+1],line=dict(color=c[wave]),mode = 'lines',
                                        legendgroup = wave, name = wave, showlegend = False))  
            i+=1
    fig.update_layout(title_text= f"{title} Interactive viewer")
    fig.update_layout( xaxis=dict( rangeslider=dict(visible=True), type="linear"),  # Add range slider
                       yaxis=dict( fixedrange = False)  )
    fig.show()

######################################################
## Illustration of signal transformation techniques ##
######################################################
from scipy.signal import find_peaks
def draw_fft_diagrams(recording, ana_file, sr = 100, wave_type: str = None, which = 0, 
                               log = False, is_average = False, n_windows = 10):
    '''
    Input:
        recording: dataframe containing wave data - 2 columns [time,amplitude]
        ana: analysis file corresponding to the recording
        wave_type: 'whole' or type of waves in 'NP', 'C', 'E1', 'E2', 'F', 'G', 'pd'
        is_average: if True, takes the average of all fft rows
        n_windows: the number of coefficients rows to plot
    Output:
        plot of the fft coefficients
    '''    
    ana = dc(ana_file)
    ana['time'] = ana['time'].apply(lambda x: x*sr)
    ana['time'] = ana['time'].astype(int)

    waveform_indices = get_index(ana)
    waveform_indices_sub = waveform_indices[wave_type]#get wave indices
    start, end = waveform_indices_sub[which]

    if end-start < 1024:
        wave_sample = recording[(start+end)//2-1024:(start+end)//2+1024]
    else:
        wave_sample = recording[start:end]

    #short-time fourier transform
    stft_coefs = np.abs(librosa.stft(wave_sample,n_fft = 1024,center = False))/np.sqrt(1024)

    if log == True:
        stft_coefs = np.log(stft_coefs)

    freq_axis = librosa.fft_frequencies(sr = sr, n_fft = 1024)

    f,(ax1,ax2) = plt.subplots(1,2,figsize=(12,2))

    if is_average == True:
        avg = np.sum(stft_coefs,axis = 1)/stft_coefs.shape[1] #average
        peaks_idx = find_peaks(avg,prominence = 0.75)[0]
        ax1.plot(freq_axis[peaks_idx], avg[peaks_idx], "xr")
        ax1.plot(freq_axis,avg,'r')
        ax1.set_title(f"log-Average of {wave_type} wave windows" if log == True else f"Average of {wave_type} wave windows")

    else: 
        for i in range(n_windows):
            ax1.plot(freq_axis, stft_coefs[:,i],color = 'r')
        ax1.set_title(f"log coefs of {n_windows} {wave_type} wave windows" if log == True else f"Frequency domain representation")
    ax1.set_xlabel('Frequency')
    ax1.set_ylabel('log - Amplitude' if log == True else 'Amplitude')

    ax2.set_xlabel('Time')
    ax2.plot(wave_sample, color = 'blue')
    ax2.set_title(f"Time domain representation")
    

###############################
## Result plotting utilities ##
###############################

def plot_pred_proba(predicted_probability, hop_length, scope, r: tuple = None, ax = None):
    '''
        Plot the predicted probability distributions
        ---------
        Argument
        ---------
            r:tuple (start,end) or None: range to plot
            ax: matplotlib.pytplot.axis object or None: axis to plot 
    '''
    test_hop_length = hop_length // scope
    pred_proba = repeat(predicted_probability, test_hop_length)
    pred_proba = extend(pred_proba, 2880000)
    
    if ax == None:
        ax = plt.gca()
    if r is not None:
        r = np.arange(r[0],r[1])
        pp = pred_proba[r]
    else:
        pp = pred_proba

    waveform = list(c.keys())
    lower = np.zeros(pp.shape[0])
    upper = np.zeros(pp.shape[0])
    x = np.arange(pp.shape[0])
    
    for i in range(len(waveform)):
        lower = upper 
        upper = upper + pp[:,i]
        plt.fill_between(x, lower, upper, alpha = 0.2, color = c[waveform[i]])
        plt.plot(x, upper, color = c[waveform[i]],linewidth = 0.4, label = waveform[i])

    plt.legend(loc = 'upper right')

def plot_gt_vs_pred_segmentation(recording, gt_ana, pred_ana = None, which = 'pred_vs_gt', savefig = False, name: str = ''): 

    if gt_ana is None:
        which = 'prediction'

    if which == 'prediction':
        plt.figure(figsize=(16,3))
        visualize_signal(recording, pred_ana)
        plt.title('Prediction')
        plt.tight_layout()
    elif which == 'ground_truth':
        assert (gt_ana is not None), 'Ground-truth analysis not found.'
        plt.figure(figsize=(16,3))
        visualize_signal(recording, gt_ana)
        plt.title('Prediction')
        plt.tight_layout() 

    elif which == 'pred_vs_gt':
        f, ax = plt.subplots(2, 1, figsize=(16,5))
        
        visualize_signal(recording, gt_ana, ax=ax[0])
        ax[0].text(0.85, 0.1, 'Ground-truth', horizontalalignment='center', verticalalignment='center', transform=ax[0].transAxes)
        ax[0].set_xlabel('')
        ax[0].set_xticks([])
        ax[0].set_title(name)
        visualize_signal(recording, pred_ana, ax = ax[1])
        ax[1].text(0.85, 0.1, 'Predicted', horizontalalignment='center', verticalalignment='center', transform=ax[1].transAxes)
        plt.subplots_adjust(hspace = 0)

        plt.tight_layout()
    else:
        raise ValueError("Param which must be one of ['prediction', 'ground_truth', 'pred_vs_gt'].")

    if savefig == True:
        os.makedirs('./prediction/figures', exist_ok = True)
        dir = os.listdir('./prediction/figures')
        index = len(dir) + 1
        if name == '':
            plt.savefig(f'./prediction/figures/Untitled{index}.png')
        else:
            plt.savefig(f'./prediction/figures/{name}.png')


#############
### Miscs ###
#############

def extend(arr1, target):
    '''
        Repeat the last axis of arr1 until 
        as long as target array (or reach target length)
    '''
    if isinstance(target, np.ndarray):
        ext_len = len(target) - len(arr1)
    else: 
        ext_len = target - len(arr1)
    if ext_len < 0:
        raise RuntimeError('Cannot repeat. Target length is less than input length')
    if len(arr1.shape) == 1: 
        extension = np.repeat(arr1[-1], ext_len)
    else:
        extension = np.tile(arr1[-1,:], (ext_len,1))
    return np.concatenate([arr1, extension])

def repeat(array, coef):
    shape = array.shape 
    if len(shape) == 3:
        if shape[1] == 1:
            tmp = array.reshape(([shape[0],shape[2]]))
    n = len(array)
    repeated = []
    for i in range(n):
        repeated.extend([array[i]] * coef)
    return np.array(repeated)


import torch
import torch.nn as nn
import torch.nn.functional as F


# Util function for computing the output size of a convolutional neuron
def compute_size(input_size, conv_layer):
    kernel_size = conv_layer.kernel_size[0]
    dilation = conv_layer.dilation[0]
    stride = conv_layer.stride[0]
    padding = conv_layer.padding[0]
    new_size = int((input_size + 2*padding - dilation*(kernel_size - 1) - 1)//stride + 1)
    return new_size

class CNN2D(nn.Module):
    def __init__(self, input_size, n_class = 7):
        torch.manual_seed(28)
        super().__init__()
        self.__type__ = 'cnn'
        self.__arch__ = 'cnn2d'
        self.__version__ = 'baseline'
        self.conv1 = nn.Conv2d(in_channels = 1, out_channels = 16, kernel_size = 3, stride = 1)
        out_size = compute_size(input_size, self.conv1) // 2
        self.conv2 = nn.Conv2d(in_channels = 16, out_channels = 16, kernel_size = 3, stride = 1)
        out_size = compute_size(out_size, self.conv1) // 2
        self.pool = nn.MaxPool2d(kernel_size=2, stride = 2)
        self.relu = nn.ReLU()
        self.dropout1 = nn.Dropout(0.3)
        self.dropout2 = nn.Dropout(0.3)
        self.fc1 = nn.Linear(16*out_size*out_size, 256)
        self.fc2 = nn.Linear(256, n_class)
        self.init_weights = False

    def forward(self,x):
        x = self.pool(self.relu(self.conv1(x)))
        x = self.pool(self.relu(self.conv2(x)))
        x = x.flatten(1)
        x = self.dropout1(x)
        x = self.relu(self.fc1(x))
        x = self.dropout2(x)
        x = self.fc2(x)
        x = F.log_softmax(x,-1)
        return x




from sklearn.metrics import accuracy_score, confusion_matrix, f1_score, precision_score, recall_score
from easydict import EasyDict
import numpy as np 


def top_k_accuracy(pred_proba, true_label, k = 2):
    n = len(true_label)
    n_class = len(pred_proba[0]) - 1
    n_correct = 0
    for i in range(n):
        top_k = np.argsort(pred_proba[i])[:n_class-k:-1]
        if true_label[i] in top_k:
            n_correct +=1 
    return n_correct/n 

def scoring(true_label, predicted_label):

    acc = np.round(accuracy_score(true_label,predicted_label),4)
    f1 = np.round(f1_score(true_label, predicted_label, average= 'macro',zero_division = 0),4)
    precision = np.round(precision_score(true_label, predicted_label, average= 'macro',zero_division = 0),4)
    recall = np.round(recall_score(true_label, predicted_label, average= 'macro',zero_division = 0),4) 
    c = np.round(confusion_matrix(true_label,predicted_label, normalize='pred'),4)
    
    return {'scores': EasyDict({'accuracy': acc, 'f1': f1, 'precision': precision, 'recall': recall}), 'confusion_matrix': c}



import numpy as np
from tqdm import tqdm
import numpy as np
import pywt
from copy import deepcopy as dc

'''
    Adapted from 
        https://github.com/uchidalab/time_series_augmentation
    
'''
def shifting(x, sigma = 1):
    return x + np.random.normal(loc = 0, scale = sigma)*np.ones_like(x)

def jitter(x, sigma=0.03):
    # https://arxiv.org/pdf/1706.00527.pdf
    return x + np.random.normal(loc=0., scale=sigma, size=(x.shape[0], x.shape[1]))

def scaling(x, sigma=0.1):
    # https://arxiv.org/pdf/1706.00527.pdf
    factor = np.random.normal(loc=1., scale=sigma, size=x.shape[0])
    return np.multiply(x.T, factor).T

def window_warp(x, window_ratio=0.1, scales=[0.5, 2.]):
    # https://halshs.archives-ouvertes.fr/halshs-01357973/document
    warp_scales = np.random.choice(scales, x.shape[0])
    warp_size = np.ceil(window_ratio*x.shape[1]).astype(int)
    window_steps = np.arange(warp_size)
        
    window_starts = np.random.randint(low=1, high=x.shape[1]-warp_size-1, size=x.shape[0]).astype(int)
    window_ends = (window_starts + warp_size).astype(int)
            
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        start_seg = pat[:window_starts[i]]
        window_seg = np.interp(np.linspace(0, warp_size-1, num=int(warp_size*warp_scales[i])), window_steps, pat[window_starts[i]:window_ends[i]])
        end_seg = pat[window_ends[i]:]
        warped = np.concatenate((start_seg, window_seg, end_seg))                
        ret[i,:] = np.interp(np.arange(x.shape[1]), np.linspace(0, x.shape[1]-1., num=warped.size), warped).T
    return ret

def rotation(x):
    flip = np.random.choice([-1, 1], size=(x.shape[0],x.shape[-1]))
    rotate_axis = np.arange(x.shape[-1])
    np.random.shuffle(rotate_axis)    
    return flip[:,np.newaxis,:] * x[:,:,rotate_axis]

def permutation(x, max_segments=5, seg_mode="equal"):
    orig_steps = np.arange(x.shape[1])
    
    num_segs = np.random.randint(1, max_segments, size=(x.shape[0]))
    
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        if num_segs[i] > 1:
            if seg_mode == "random":
                split_points = np.random.choice(x.shape[1]-2, num_segs[i]-1, replace=False)
                split_points.sort()
                splits = np.split(orig_steps, split_points)
            else:
                splits = np.array_split(orig_steps, num_segs[i])
            warp = np.concatenate(np.random.permutation(splits)).ravel()
            ret[i] = pat[warp]
        else:
            ret[i] = pat
    return ret

def magnitude_warp(x, sigma=0.2, knot=4):
    from scipy.interpolate import CubicSpline
    orig_steps = np.arange(x.shape[1])
    
    random_warps = np.random.normal(loc=1.0, scale=sigma, size=(x.shape[0], knot+2, x.shape[-1]))
    warp_steps = (np.ones((x.shape[-1],1))*(np.linspace(0, x.shape[1]-1., num=knot+2))).T
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        warper = np.array([CubicSpline(warp_steps[:,dim], random_warps[i,:,dim])(orig_steps) for dim in range(x.shape[-1])]).T
        ret[i] = pat * warper

    return ret

def time_warp(x, sigma=0.2, knot=4):
    from scipy.interpolate import CubicSpline
    orig_steps = np.arange(x.shape[1])
    
    random_warps = np.random.normal(loc=1.0, scale=sigma, size=(x.shape[0], knot+2, x.shape[-1]))
    warp_steps = (np.ones((x.shape[-1],1))*(np.linspace(0, x.shape[1]-1., num=knot+2))).T
    
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[-1]):
            time_warp = CubicSpline(warp_steps[:,dim], warp_steps[:,dim] * random_warps[i,:,dim])(orig_steps)
            scale = (x.shape[1]-1)/time_warp[-1]
            ret[i,:,dim] = np.interp(orig_steps, np.clip(scale*time_warp, 0, x.shape[1]-1), pat[:,dim]).T
    return ret



# PREPROCESSING
def downsampling(wave_array, coef = 10):
    '''
        Downsample the signal by average 
    '''
    new_length = len(wave_array)//coef
    downsampled_wave = []
    for i in range(new_length):
        downsampled_wave.append(np.mean(wave_array[i*10:(i+1)*10]))
    return np.array(downsampled_wave)

def quantile_filter(wave_array, k = 1.5, frame_length = 10, sr = 100):
    '''
    Input:
        wave: wave signal of class np.ndarray
    Output:
        new signal whose outliers are set to zeros
    '''
    copy_array = dc(wave_array)
    n_frame = len(wave_array)//(frame_length*sr)
    for i in range(n_frame):
        idx = np.arange(i*frame_length*sr, (i+1)*frame_length*sr)
        w = copy_array[idx]
        Q1 = np.quantile(w,0.25)
        Q3 = np.quantile(w,0.75)
        IQR = abs(Q1-Q3) 
        for i in range(len(w)):
            if (w[i] <= Q1-k*IQR):
                w[i] = Q1-k*IQR
            elif (w[i] >= Q3+k*IQR):
                w[i] = Q3+k*IQR
        copy_array[idx] = w
    return copy_array

"""
Wavelet transforms
"""
def wavelet_reconstruction(wave, wavelet, level, threshold = 'global'):

    coeffs = pywt.wavedec(wave, wavelet, level=level)
    all_coeffs = np.concatenate([l for l in coeffs])

    if threshold == 'global':
        N = len(wave)
        sigma = np.median(np.abs(all_coeffs))/0.6745
        t = sigma*np.sqrt(2*np.log(N))

    elif threshold == 'std':
        std = np.std(all_coeffs,ddof = 1)
        t = 1.5*std
        
    elif threshold == 'ratio':
        t = 0.05*np.max(all_coeffs)
    
    else:
        t = threshold
        
    coeffs_thresholded = [pywt.threshold(c, t, mode='soft') for c in coeffs]
    denoised = pywt.waverec(coeffs_thresholded, wavelet)

    return denoised



from setuptools import find_packages, setup 

with open('package/Introduction.md', 'r', encoding="utf-8") as f:
    long_description = f.read()

setup(
    name = 'DiscoEPG',
    version = '0.0.6',
    description = "Open-source package for insects EPG data analysis",
    package_dir = {"":"package"},
    packages= find_packages(where = "package"),
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = "https://github.com/HySonLab/ML4Insects",
    author = "DINHQuangDung1999",
    author_email = "dqdung999@gmail.com",
    license="MIT",
    classifiers= [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent" 
    ],
    extra_requires = {
        "twine >= 4.0.2"
    },
    python_requires = ">=3.10"
)


# MIT License
# 
# Copyright (c) 2018 Tom Runia
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-04-16

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import torch
import torch.nn as nn


class TorchFilterBank(nn.Module):

    def __init__(self, filters=None, cuda=True):
        """
        Temporal filter bank in PyTorch storing a collection of nn.Conv1d filters.
        When cuda=True, the convolutions are performed on the GPU. If initialized with
        filters=None, the set_filters() method has to be called before actual running
        the convolutions.

        :param filters: list, collection of variable sized 1D filters (default: [])
        :param cuda: boolean, whether to run on GPU or not (default: True)
        """
        super(TorchFilterBank, self).__init__()
        self._cuda = cuda
        self._filters = [] if not filters else self.set_filters(filters)

    def forward(self, x):
        """
        Takes a batch of signals and convoles each signal with all elements in the filter
        bank. After convoling the entire filter bank, the method returns a tensor of
        shape [N,N_scales,1/2,T] where the 1/2 number of channels depends on whether
        the filter bank is composed of real or complex filters. If the filters are
        complex the 2 channels represent [real, imag] parts.

        :param x: torch.Variable, batch of input signals of shape [N,1,T]
        :return: torch.Variable, batch of outputs of size [N,N_scales,1/2,T]
        """

        if not self._filters:
            raise ValueError('PyTorch filters not initialized. Please call set_filters() first.')
            return None
        results = [None]*len(self._filters)
        for ind, conv in enumerate(self._filters):
            results[ind] = conv(x)
        results = torch.stack(results)     # [n_scales,n_batch,2,t]
        results = results.permute(1,0,2,3) # [n_batch,n_scales,2,t]
        return results

    def set_filters(self, filters, padding_type='SAME'):
        """
        Given a list of temporal 1D filters of variable size, this method creates a
        list of nn.conv1d objects that collectively form the filter bank.

        :param filters: list, collection of filters each a np.ndarray
        :param padding_type: str, should be SAME or VALID
        :return:
        """

        assert isinstance(filters, list)
        assert padding_type in ['SAME', 'VALID']

        self._filters = [None]*len(filters)
        for ind, filt in enumerate(filters):

            assert filt.dtype in (np.float32, np.float64, np.complex64, np.complex128)

            if np.iscomplex(filt).any():
                chn_out = 2
                filt_weights = np.asarray([np.real(filt), np.imag(filt)], np.float32)
            else:
                chn_out = 1
                filt_weights = filt.astype(np.float32)[None,:]

            filt_weights = np.expand_dims(filt_weights, 1)  # append chn_in dimension
            filt_size = filt_weights.shape[-1]              # filter length
            padding = self._get_padding(padding_type, filt_size)

            conv = nn.Conv1d(1, chn_out, kernel_size=filt_size, padding=padding, bias=False)
            conv.weight.data = torch.from_numpy(filt_weights)
            conv.weight.requires_grad_(False)

            if self._cuda: conv.cuda()
            self._filters[ind] = conv

    @staticmethod
    def _get_padding(padding_type, kernel_size):
        assert isinstance(kernel_size, int)
        assert padding_type in ['SAME', 'VALID']
        if padding_type == 'SAME':
            return (kernel_size - 1) // 2
        return 0